from abc import abstractmethod
from typing import Generator

from ..utils import DataContainer


class Source:

    def __init__(self, name: str, **kwargs):
        self.name = name
        for key, value in kwargs.items():
            setattr(self, key, value)

    @abstractmethod
    def get(self, **kwargs) -> Generator[any]:
        raise NotImplementedError

    @abstractmethod
    def adapter(self, data, **kwargs) -> Generator[DataContainer]:
        raise NotImplementedError

    def set_params(self, **kwargs) -> None:
        for key, value in kwargs.items():
            setattr(self, key, value)

    def get_params(self) -> dict[str, any]:
        return {key: value for key, value in vars(self).items() if key != "name"}



import os
from typing import override

from ..utils.exceptions import UnsupportedFileType
from . import Destination
from ..utils import DataContainer
from pathlib import Path
import re
import boto3
import tempfile


class s3(Destination):

    def __init__(
        self, name: str, endpoint: str, access_key_id: str, secret_access_key: str, bucket_name: str, folder: str
    ):
        super().__init__(name)
        self.bucket = bucket_name
        self.folder = folder
        url = "https://" + re.sub(r"^(https?://)?", "", endpoint)
        self._s3_client = boto3.client(
            "s3", endpoint_url=url, aws_access_key_id=access_key_id, aws_secret_access_key=secret_access_key
        )
        self._tmp_dirname = Path(tempfile.mkdtemp())

    @override
    def dump(self, data: DataContainer) -> None:
        file_format = data.get_format() if data.get_format() else "csv"
        df = data.get_data()
        df.to_csv(data.get_name() + file_format)

        file_name = data.get_name() + "." + file_format
        output_path = os.path.join(self._tmp_dirname, file_name)
        if file_format == "csv":
            df.to_csv(output_path, sep=";")
        elif file_format == "json":
            df.to_json(output_path)
        elif file_format == "xml":
            df.to_excel(output_path)
        elif file_format == "xlsx":
            df.to_excel(output_path)
        elif file_format == "parquet":
            df.to_parquet(output_path)
        else:
            raise UnsupportedFileType(f"{file_format} format is not supported at the moment.")

        self._s3_client.upload_file(output_path, self.bucket, os.path.join(self.folder, file_name))



from typing import override

from sqlalchemy import create_engine

from . import Destination
from ..utils import DataContainer


class Postgres(Destination):

    def __init__(self, name: str, host: str, port: int, database: str, user: str, schema: str, password: str):
        super().__init__(name)
        self.host = host
        self.port = port
        self.database = database
        self.schema = schema
        self.user = user
        self.password = password
        self.engine = create_engine(
            f"postgresql://{self.user}:{self.password}@{self.host}:{self.port}/{self.database}?sslmode=require"
        )

    @override
    def dump(self, data: DataContainer) -> None:
        table_name = data.get_name()
        df = data.get_data()
        df.to_sql(table_name, self.engine, schema=self.schema, index=False, if_exists="append", method="multi")



import importlib
import inspect
import json
import os
from pathlib import Path
from typing import Literal

import yaml
from pydantic import BaseModel

from .exceptions import ModelNotFound


def load_config(config_file_path: str, step: Literal["source", "destination"], name: str) -> BaseModel:
    models = load_models_from_module(os.path.join("etl/configs/", step))
    pydantic_model = models.get(name.lower())
    if not pydantic_model:
        raise ModelNotFound(f"'{name}' not found in {list(models.keys())}")

    file_extension = config_file_path.split(".")[-1]
    with open(config_file_path, encoding="utf8") as file:
        if file_extension == "json":
            config_data = json.load(file)
        elif file_extension == "yaml":
            config_data = yaml.safe_load(file)
        else:
            raise ValueError("Unsupported config format, use json or yaml.")

    return pydantic_model(**config_data)


# Function to dynamically load models from Python files in a submodule
def load_models_from_module(module_path: str) -> dict[str, type]:
    model_map = {}

    # Traverse the submodule directory and import each Python file
    module_dir = Path(module_path)

    if not module_dir.is_dir():
        raise ValueError(f"The provided path {module_path} is not a directory")

    # Iterate over all Python files in the directory
    for py_file in module_dir.glob("*.py"):

        # Get the module name from the file path
        module_name = py_file.stem

        # Dynamically import the module
        module = importlib.import_module(f"{module_path.replace('/', '.')}.{module_name}")

        # Iterate over all classes in the module and check if they are Pydantic models
        for name, obj in inspect.getmembers(module, inspect.isclass):
            if issubclass(obj, BaseModel) and obj is not BaseModel:
                model_map[name.lower()] = obj

    return model_map



class UnsupportedFileType(Exception):
    pass


class UntrustedSource(Exception):
    pass


class MissingParameter(Exception):
    pass


class InapplicableTransformation(Exception):
    pass


class AdapterConversion(Exception):
    pass


class ModelNotFound(Exception):
    pass



import re
import warnings
from io import BytesIO

import pandas as pd
from pandas import DataFrame


class DataContainer:

    def __init__(self, name: str, data: DataFrame, fmt: str = ""):
        cleaned_name = re.sub(r"[^a-zA-Z0-9]", "_", name)
        normalized_name = re.sub(r"_+", "_", cleaned_name).rstrip("_").lstrip("_")
        self._name = normalized_name.lower()
        if not self._name[0].isalpha():
            warnings.warn(f"The name '{self._name}' does not start with a letter. This may cause issues.", stacklevel=2)
        self._data = self._serialize_data(data)
        self._format = fmt

    @staticmethod
    def _serialize_data(data: DataFrame) -> bytes:
        buffer = BytesIO()
        data.to_parquet(buffer, index=False)
        return buffer.getvalue()

    @staticmethod
    def _deserialize_data(data_bytes: bytes) -> DataFrame:
        buffer = BytesIO(data_bytes)
        return pd.read_parquet(buffer)

    def get_name(self) -> str:
        return self._name

    def get_data(self) -> DataFrame:
        return self._deserialize_data(self._data)

    def get_format(self) -> str | None:
        return None if self._format == "" else self._format



import inspect
from datetime import datetime
from typing import Callable

import pandas as pd
from loguru import logger
import os

from ..destination import Destination
from ..source import Source
from ..utils.exceptions import InapplicableTransformation, AdapterConversion

log_level = os.environ.get("LOG_LEVEL") or "DEBUG"
logger.level(log_level)


class Pipeline:

    def __init__(self, pipeline_name: str, source: Source, destination: Destination) -> None:
        self.name = pipeline_name
        self.logger = logger
        self.source = source
        self.destination = destination
        self.transfo: list[Callable[[pd.DataFrame], pd.DataFrame]] = []

    def run(self) -> None:
        starting_time = datetime.now()
        self._info(f"STARTING PIPELINE {self.name}")
        self._debug(f"Starting time {starting_time}")

        self._info(f"LOADING DATA FROM SOURCE {self.source.name} OF TYPE : {self.source.__class__}")
        load_data = self.source.get()
        if load_data is None:
            self._warn("No data found closing pipeline")
            return
        self._debug("Data fetched successfully")

        self._info(f"ADAPTING DATA FROM SOURCE {self.source.name} OF TYPE : {self.source.__class__}")
        data = self.source.adapter(load_data)
        if data is None:
            error_msg = "The source adapter returned no data while data were fetched from the source."
            self._error(error_msg)
            raise AdapterConversion(error_msg)

        self._info("APPLYING CUSTOM FUNCTION TO ADAPTED DATA")
        for transfo in self.transfo:
            try:
                self._debug(f"Data :\n{data.describe()}")
                self._debug(f"Function :\n{inspect.getsource(transfo)}")
                data = transfo(data)
            except BaseException:  # noqa:B036
                error_message = f"Cannot apply custom transformation for function :\n{inspect.getsource(transfo)}"
                self._error(error_message)
                raise InapplicableTransformation(error_message)

        self._info(f"DUMPING DATA TO DESTINATION {self.destination.name} OF TYPE {self.destination.__class__}")
        for d in data:
            self.destination.dump(d)

        self._info("CLOSING PIPELINE")
        ending_time = datetime.now()
        self._debug(f"Ending time : {ending_time}")
        runtime = ending_time - starting_time
        self._debug(f"Runtime : {runtime}")

    def add_transfo(self, transfo: Callable[[pd.DataFrame], pd.DataFrame]) -> None:
        self.transfo.append(transfo)

    def _debug(self, msg: str) -> None:
        self.logger.debug(self._enriched_log_msg(msg))

    def _info(self, msg: str) -> None:
        enriched_msg = self._enriched_log_msg(msg)
        upper_msg = enriched_msg.upper()
        spaced_msg = " " + upper_msg + " "
        centered_msg = spaced_msg.center(100, "*")
        self.logger.info(centered_msg)

    def _warn(self, msg: str) -> None:
        self.logger.warning(self._enriched_log_msg(msg))

    def _error(self, msg: str) -> None:
        self.logger.error(self._enriched_log_msg(msg))

    def _enriched_log_msg(self, msg: str) -> str:
        enriched_msg = f"PIPELINE {self.name} - " + msg
        return enriched_msg



import os
import re
import tempfile
from pathlib import Path
from typing import override, Generator, Union

import boto3
import pandas as pd

from . import Source
from ..utils import DataContainer
from ..utils.exceptions import UnsupportedFileType, UntrustedSource


class S3(Source):

    def __init__(self, endpoint: str, access_key_id: str, secret_access_key: str, bucket_name: str):
        super().__init__()
        self.bucket = bucket_name
        url = "https://" + re.sub(r"^(https?://)?", "", endpoint)
        self._s3_client = boto3.client(
            "s3", endpoint_url=url, aws_access_key_id=access_key_id, aws_secret_access_key=secret_access_key
        )
        self._tmp_dirname = Path(tempfile.mkdtemp())

    @override
    def get(self, incremental: str = None, folder: str = "") -> Generator[Union[str, bytes, os.PathLike]]:
        s3_file_paths = self._list_folder(folder)
        file_names = [os.path.basename(file_path) for file_path in s3_file_paths]
        for file_name, s3_path in zip(file_names, s3_file_paths):
            local_file_path = os.path.join(self._tmp_dirname, file_name)
            self._download_file(s3_path, local_file_path)
            yield local_file_path

    @override
    def adapter(self, data: Generator[Union[str, bytes, os.PathLike]], **kwargs) -> Generator[DataContainer]:
        for d in data:
            file_type = os.path.basename(d).split(".")[-1].lower()
            file_name = os.path.basename(d).split(".")[0].lower()
            if file_type == "csv":
                yield DataContainer(file_name, pd.read_csv(d, header=True, **kwargs), "csv")
            elif file_type == "json":
                yield DataContainer(file_name, pd.read_json(d, **kwargs), "json")
            elif file_type == "html":
                for count, html_data in enumerate(pd.read_html(d, **kwargs)):
                    yield DataContainer(f"{count}_{file_name}", html_data, "html")
            elif file_type == "xml":
                yield DataContainer(file_name, pd.read_xml(d, **kwargs), "xml")
            elif file_type == "xlsx":
                yield DataContainer(file_name, pd.read_excel(d, **kwargs), "xlsx")
            elif file_type == "parquet":
                yield DataContainer(file_name, pd.read_parquet(d, **kwargs), "parquet")
            elif file_type == "pickle":
                raise UntrustedSource("Pickle is not a safe source to use and can lead to security breach.")
            else:
                raise UnsupportedFileType(f"{file_type} format is not supported at the moment.")

    def _list_folder(self, folder) -> list[str]:
        response = self._s3_client.list_objects(Bucket=self.bucket, Prefix=folder)
        if "Contents" not in response.keys():
            return []
        files = [file.get("Key") for file in response.get("Contents") if response.get("Key") != folder]
        return files

    def _download_file(self, target_filename: str, output_filename: str) -> None:
        self._s3_client.download_file(self.bucket, target_filename, output_filename)
